/******************************************************************************
	adtek_8132_I_06.c 
	2001/05/18	by m.tadano
	cc -c -g -Wall -O -v adtek_8132_I_06.c -o adtek_8132.o -DDDRIVER
	# mknod -m 666 /dev/adtek_8132 c 251 0
	# insmod adtek_8132.o major=251	(set major number FIX assignment)
	# insmod adtek_8132.o		(set major number AUTO assignment)
	# insmod adtek_8132.o major=251 irq_enable=0x1
				irq_enable : 0x1, 0x10, 0x100 and/or 0x1000
******************************************************************************/

#ifndef	MODULE
#define	MODULE
#endif
#ifndef	__KERNEL__
#define	__KERNEL__
#endif

#include <linux/version.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/signal.h>
#include <linux/major.h>
#include <linux/module.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/poll.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/malloc.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "adtek_8132_I.h"

static unsigned int	busy = 0;
static struct pci_dev	*pci_device = NULL;
static int	interrupt_req = 0;
u8	irq_set;

static dDriver_t dDriver;

static int	major = 0;
static int	irq_enable = 0;
static int	irq_type = TRG_MODE_EDGE;
static int	irq_polarity = TRG_LOGIC_HIGH;
MODULE_PARM (major, "i");
MODULE_PARM (irq_enable, "i");
MODULE_PARM (irq_type, "i");
MODULE_PARM (irq_polarity, "i");

/* interrupt *****************************************************************/
static void	adtek_8132_interrupt (
			int irq, void *dev_id, struct pt_regs *regs)
{
int	flag_int;

	flag_int = inb (ADR_INT_FLAG);
	if (0 < flag_int)	{
#ifdef DEBUG
		printk ("ADR_INT_FLAG(interrupt) : %x\n", flag_int);
#endif
		if (irq_polarity == TRG_LOGIC_HIGH)
			irq_polarity = TRG_LOGIC_LOW;
		else
			irq_polarity = TRG_LOGIC_HIGH;
		irq_set = irq_type + (irq_enable & 0xff) + irq_polarity;
		outb (irq_set, ADR_INT_CNT1);	/* interrupt register set */
		irq_set = irq_type + ((irq_enable & 0xff00) >> 8) + irq_polarity;
		outb (irq_set, ADR_INT_CNT2);
		outb (flag_int, ADR_INT_CLEAR);
		interrupt_req = flag_int;
		wake_up_interruptible (&dDriver.queue);
	}
}

/* poll **********************************************************************/
static unsigned int adtek_8132_poll (struct file *filep, poll_table *wait)
{

#ifdef DEBUG
	printk ("adtek_8132_poll/ ");
#endif
	poll_wait (filep, &dDriver.queue, wait);
	if (interrupt_req == ON)	{
		interrupt_req = OFF;
		return POLLIN | POLLRDNORM;
	}
#ifdef DEBUG
	printk ("adtek_8132_poll/ return\n ");
#endif
	return 0;
}

/* open **********************************************************************/
static int adtek_8132_open (struct inode *inode, struct file *file)
{
int	minor;
int	result;

	printk ("adtek_8132_open\n");
	if (busy != 0) 
		return - EBUSY;
	minor = MINOR (inode->i_rdev);
	if (minor < 0)
  		return -ENODEV;
	if (irq_enable != 0)	{
		if (result = (request_irq (dDriver.irq, adtek_8132_interrupt, 
			SA_INTERRUPT | SA_SHIRQ, DEVICE_NAME, pci_device)))	{
			printk ("adtek_8132_open:request_irq error: error=%d\n", 
								result);
		return result;
		}
	}
	init_waitqueue_head (&dDriver.queue);
	irq_set = irq_type + (irq_enable & 0xff) + irq_polarity;
	outb (irq_set, ADR_INT_CNT1);		/* interrupt register set */
	irq_set = irq_type + ((irq_enable & 0xff00) >> 8) + irq_polarity;
	outb (irq_set, ADR_INT_CNT2);
	dDriver.minor = minor;
	busy = 1;
	MOD_INC_USE_COUNT;

	printk ("adtek_8132_open/ return\n");
	return 0;
}

/* close *********************************************************************/
static int adtek_8132_close (struct inode *inode, struct file *file)
{
	printk ("adtek_8132_close\n");

	busy = 0;
	if (irq_enable != 0)
		free_irq(dDriver.irq, pci_device);
	MOD_DEC_USE_COUNT;
	
	printk ("adtek_8132_close/ return\n");
	return 0;
}

/* read **********************************************************************/
static ssize_t adtek_8132_read (
		struct file *file, char *buf, size_t count, loff_t *ppos)
{
static unsigned char	io_buf[8];
int	i;

#ifdef DEBUG
	printk ("adtek_8132_read/ count = %d ", count);
#endif
	for (i = 0; i < MAX_PORT; i++)
		io_buf[i] = inb (dDriver.iobase1+i);
	if (copy_to_user (buf, io_buf, MAX_PORT) != 0)	{
		printk ("adtek_8132_read/ ERROR !/");
		return -EFAULT;
	}

#ifdef DEBUG
	printk ("adtek_8132_read/ return\n");
#endif
	return 0;
}


/* file_operations *************************************************************/
static struct file_operations adtek_8132_fops =
{
  owner:	THIS_MODULE,
  read:		adtek_8132_read,
  poll:		adtek_8132_poll,
  open:		adtek_8132_open,
  release:	adtek_8132_close,	
};

/* init_module *****************************************************************/
int init_module (void)
{
struct pci_dev	*ret_dev;
int	result;

	printk ("init_module/adtek_8132/ ");
	ret_dev = pci_find_device (ADTEK_8132_IN_BOARD_VENDOR,
                              ADTEK_8132_IN_BOARD_DEVICE, pci_device);
	if (!ret_dev)
		return -ENODEV;         /* not found */
	pci_device = ret_dev;
	result = pci_enable_device (pci_device);
	if (result)
		return result;
	dDriver.iobase0 = pci_resource_start (pci_device, 0);
	dDriver.iobase1 = pci_resource_start (pci_device, 1);
	dDriver.iobase2 = pci_resource_start (pci_device, 2);
	dDriver.iobase3 = pci_resource_start (pci_device, 3);
	dDriver.iobase4 = pci_resource_start (pci_device, 4);
	dDriver.iobase5 = pci_resource_start (pci_device, 5);
	dDriver.irq = pci_device ->irq;
	request_region (dDriver.iobase0, 16, DEVICE_NAME);
	request_region (dDriver.iobase1, 16, DEVICE_NAME);

	result = register_chrdev (major, DEVICE_NAME, &adtek_8132_fops);
	if (result < 0)	{
		unregister_chrdev (major, DEVICE_NAME);
		return - EIO;
	}
	if (major == 0)
	  	dDriver.major = result;
	else
		dDriver.major = major;
	
	printk ("iobase0=%x ", dDriver.iobase0);
	printk ("iobase1=%x ", dDriver.iobase1);
	printk ("iobase2=%x ", dDriver.iobase2);
	printk ("iobase3=%x ", dDriver.iobase3);
	printk ("iobase4=%x ", dDriver.iobase4);
	printk ("iobase5=%x ", dDriver.iobase5);
	printk ("irq=%d \n", dDriver.irq);
	printk ("irq_enable=%x irq_type=%x irq_polarity=%x\n",
			irq_enable, irq_type, irq_polarity);

	printk ("init_module/adtek_8132/end\n");
	return 0;
}
/* *****************************************************************************/
void cleanup_module (void)
{
	printk ("cleanup_module/adtek_8132/ \n");
	release_region (dDriver.iobase0, 16);
	release_region (dDriver.iobase1, 16);
	unregister_chrdev (dDriver.major, DEVICE_NAME);
	printk ("cleanup_module/adtek_8132/end\n");
}
/* end of file *****************************************************************/
