/******************************************************************************
	advanet_2610_AD08.c
	2001/04/19	by M.Tadano

	cc -c -g -v -Wall -O advanet_2610_AD07.c -o advanet_2610.o \
		-DDDRIVER

	# mknod -m 666 /dev/advanet_2610 c 253 0
	# insmod advanet_2610.o major=253 (set major number FIX assignment)
	# insmod advanet_2610.o           (set major number AUTO assignment)
******************************************************************************/

#ifndef	MODULE
#define	MODULE
#endif
#ifndef	__KERNEL__
#define	__KERNEL__
#endif

#include <linux/version.h>
#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/wait.h>
#include <linux/errno.h>
#include <linux/types.h>
#include <linux/signal.h>
#include <linux/major.h>
#include <linux/module.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/poll.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/malloc.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/uaccess.h>

#include "advanet_2610_AD.h"

void	init_2610_board(void);
static unsigned int	eeprom_read(unsigned int);
static void	wait_1ms (int);
static unsigned int	busy = 0;
static int	maxdata_count = 0;
static u16	sts_data;
static struct pci_dev	*pci_device = NULL;
static unsigned char	*kbuf = NULL;
static unsigned short	*read_buf = NULL;
static short	f_meas = OFF, int_wait = OFF;
static int	read_p, write_p;

static dDriver_t dDriver;
static set_up_t *setup;

static int	major = 0;
MODULE_PARM (major, "i");

/* interrupt *****************************************************************/
static void advanet_2610_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
int	flag_int, i;

	flag_int = inw (INT_PRIMARY);
	if ((flag_int & 0x01) == 0x01)	{
		outw (0x1, INT_PRIMARY);		/* AD int clear */
		sts_data = inw (STS_ADR);
		if ((sts_data & 0x08) == 0x08)		/* FIFO empty ? */
			return;
		do	{
			*read_buf = inw (DATA_ADR);
			sts_data = inw (STS_ADR);
			if ((sts_data & 0x08) == 0x08)
				return;
		} while  ((sts_data & 0x40) != 0x40);	/* check chn0 ? */
		for (i = 0; i < CHANNEL_MAX -1; i++)
			*(++read_buf) = inw (DATA_ADR);
		write_p++;
		read_buf++;
		if (int_wait == ON)
			wake_up_interruptible (&dDriver.queue);
	}
	else if ((flag_int & 0x02) == 0x02)	{
		outw (0x2, INT_PRIMARY);		/* FIFO harf clear */
		printk ("2610_interrupt/ FIFO harf\n");	/* why path here? */
	}
	else if ((flag_int & 0x04) == 0x04)	{
		outw (0x4, INT_PRIMARY);		/* FIFO full clear */
		printk ("2610_interrupt/ FIFO full\n");	/* why path here? */
	}
	else
		printk ("advanet_2610_interrupt/IntO/");

	return;
}

/* open **********************************************************************/
static int advanet_2610_open (struct inode *inode, struct file *file)
{
int	minor;
int	result;

	if (busy != 0) 
		return - EBUSY;
	init_2610_board();
	minor = MINOR (inode->i_rdev);
	if (minor < 0)
  		return -ENODEV;
	busy = 1;
	MOD_INC_USE_COUNT;
	int_wait = OFF;
	dDriver.minor = minor;
	if (result = (request_irq (dDriver.irq, advanet_2610_interrupt,
		SA_INTERRUPT | SA_SHIRQ, DEVICE_NAME, pci_device)))	{
		printk ("2610_open: request_irq error: error = %d\n", 
								result);
		return result;
	}
	return 0;
}

/* close *********************************************************************/
static int advanet_2610_close (struct inode *inode, struct file *file)
{

	f_meas = OFF;
	init_2610_board();
	if (kbuf != NULL)	{
		kfree (kbuf);
		kbuf = NULL;
	}
	free_irq(dDriver.irq, pci_device);
	busy = 0;
	MOD_DEC_USE_COUNT;
	
	return 0;
}

/* read **********************************************************************/
static ssize_t advanet_2610_read 
	(struct file *file, char *buf, size_t count, loff_t *ppos)
{
unsigned long	len;
int	data_count;
static unsigned long	flags;

	if (f_meas != ON)	{
		printk ("ADc Not Start!!!!\n");
		return 0;
	}
	data_count = setup->count;

	do	{
		if (write_p < data_count)	{
			int_wait = ON;
			interruptible_sleep_on (&dDriver.queue);
			int_wait = OFF;
		}
		if ((setup->mode == CONTINUAS_SINGLECHN) || 
				(setup->mode == CONTINUAS_MULTICHN))
			break;
	} while (write_p < data_count);
	if ((setup->mode != CONTINUAS_MULTICHN) && 
				(setup->mode != CONTINUAS_SINGLECHN))	{
		outw (0x00, INT_ENABLE);	/* inhibit interrupt */
		outw (0x07, INT_PRIMARY);	/* ALL int clear */
		outw (0x00, TRG_MODE);		/* Diff & Soft trigger mode */
		f_meas = OFF;
	}
	save_flags (flags);
	cli();
	len = setup->count*2*CHANNEL_MAX;
	if (copy_to_user (buf, kbuf, len) != 0)	{
		printk ("2610_read/ copy_to_user ERROR !!!!\n");
		return -EFAULT;
	}
	write_p = 0;
	read_buf = (unsigned short *)kbuf;
	restore_flags (flags);

	return len/2;
}

/* write ********************************************************************/
static ssize_t advanet_2610_write (
	struct file *file, const char *buf, size_t count, loff_t *ppos)
{
long int	interval;
	
	if (f_meas == ON)
		init_2610_board();
	setup = (set_up_t *)buf;
	if (kbuf != NULL)	{
		kfree(kbuf);
		kbuf = NULL;
	}
	interval = setup->interval;
	if (interval < 500)			/* less than 500[usec] */
		interval = 500;
	interval = (int)(interval /500);
	if ((setup->mode == CONTINUAS_SINGLECHN) || 
			(setup->mode == CONTINUAS_MULTICHN))
		setup->count = 100;
	maxdata_count = setup->count * CHANNEL_MAX;
	kbuf = kmalloc (maxdata_count, GFP_KERNEL);
/*	kbuf = kmalloc (maxdata_count, GFP_ATOMIC);
*/
	if (kbuf == NULL)	{
		printk ("2610_write/ kmalloc error\n");
		return -EFAULT;
	}
	read_buf = (unsigned short *)kbuf;
	outw (0x00, TRG_MODE);			/* soft trigger mode set */
	do {
		sts_data = inw (STS_ADR);
	} while ((sts_data & 0x4) == 0x4);	/* AD stop ? */
	if ((sts_data & 0x8) != 0x08)
		outw (0x00, FIFO_ADR);		/* FIFO clear */
	outw ((u16)interval, CLK_DATA);		/* timer set */
	outw (0x07, INT_PRIMARY);		/* ALL int clear */
	outw (0x07, INT_ENABLE);		/* Int enable set */

	switch (setup->mode)	{
	case SINGLE_SINGLECHN:
	case SINGLE_MULTICHN:
		outw (0x01, SOFT_TRG);		/* ADC one-time scan */
		f_meas = ON;
		setup->count = 1;
		break;
	case CONTINUAS_SINGLECHN:
	case CONTINUAS_MULTICHN:
		setup->count = 1;
	case MULTI_SINGLECHN:
	case MULTI_MULTICHN:
		outw (0x02, TRG_MODE);		/* pacer clock AD start*/
		f_meas = ON;
		break;
	case ADC_STOP:
		outw (0x00, TRG_MODE);		/* soft trigger mode set */
		return 0;			/* ADC Stop */
		break;
	default:
		break;
	}
	write_p = 0;
	read_p = 0;
	return 0;
}

/* ioctl *********************************************************************/
static int advanet_2610_ioctl(struct inode *inode, struct file *file,
			unsigned int cmd, unsigned long arg)
{
int	data;

	if ((EEPROM_READ <= cmd) && (cmd <= EEPROM_READ+11))	{
		data = eeprom_read (cmd - EEPROM_READ) & 0xffff;
		if (copy_to_user ((int *)arg, &data, sizeof (int)) != 0)	{
			printk ("2610_ioctl/ copy_to_user ERROR !!!!\n");
			return -EFAULT;
		}
	}
	return 0;
}
/* eeprom_read ** ????????????????? not get value ???????????????????? *******/
static unsigned int	eeprom_read(unsigned int cmd)
{
int	i, len;
char	dataval;
unsigned int	retval;
	
	cmd = ((0x06 << 6) | cmd) << 16;
	len = 25;
	retval = 0;
	outb (0x03, EEPROM_ADR);
	wait_1ms (2);
	for (i = len; len == 0; len--)	{
		dataval = (cmd & (1L << len)) ? 0x06:0x02;
		outb (dataval, EEPROM_ADR);
		wait_1ms (2);
		outb ((dataval | 0x01), EEPROM_ADR);
		wait_1ms (2);
		retval = (retval << 1) | ((inb (EEPROM_ADR) & 0x08) ? 1 : 0);
	}
	outb ( 0x00, EEPROM_ADR);
	wait_1ms (2);

	return retval;
}
/* wait_1ms ******* dummey read at status register ***************************/
static void	wait_1ms (int count)
{
int	i,j;

	for (i = 0; i < count; i++)
		for (j = 0; j < 2000; j++)
			sts_data = inw (STS_ADR);
}

/* init_board ****************************************************************/
void init_2610_board()
{
	outw (0x00, INT_ENABLE);		/* inhibit interrupt */
	outw (0x00, TRG_MODE);			/* soft trigger mode set */
	do {
		sts_data = inw (STS_ADR);
	} while ((sts_data & 0x04) == 0x04);	/* AD stop ? */
	if ((sts_data & 0x08) != 0x08)		/* FIFO empty ? */
		outw (0x00, FIFO_ADR);		/* FIFO clear */
	outw (0x07, INT_PRIMARY);		/* ALL int clear */
	f_meas = OFF;

#ifdef DEBUG
	sts_data = inw (STS_ADR);
printk ("init_2610_board/ status (0x00)=%x ", sts_data);
	sts_data = inw (INT_ENABLE);
printk ("interrupt enable (0x04)=%x ", sts_data);
	sts_data = inw (INT_PRIMARY);
printk ("interrupt primary (0x08)=%x ", sts_data);
	sts_data = inw (CLK_DATA);
printk ("pacer clock set (0x0c)=%x ", sts_data);
	sts_data = inw (EEPROM_ADR);
printk ("EEPROM (0x20)=%x\n", sts_data);
#endif
}
/* file operation ************************************************************/
static struct file_operations advanet_2610_fops =
{
  owner:	THIS_MODULE,
  read:		advanet_2610_read,
  write:	advanet_2610_write,
  ioctl:	advanet_2610_ioctl,
  open:		advanet_2610_open,
  release:	advanet_2610_close,
};

/* init **********************************************************************/
int init_module (void)
{
struct pci_dev	*ret_dev;
int	result;

	printk ("init_module/advanet_2610/ ");
	ret_dev = pci_find_device (ADVANET_2610_ADC_BOARD_VENDOR,
                              ADVANET_2610_ADC_BOARD_DEVICE, pci_device);
	if (!ret_dev)
		return -ENODEV;         /* not found */
	pci_device = ret_dev;
	result = pci_enable_device (pci_device);
	if (result)
		return result;

	dDriver.iobase0 = pci_resource_start (pci_device, 0);
	dDriver.iobase1 = pci_resource_start (pci_device, 1);
	dDriver.iobase2 = pci_resource_start (pci_device, 2);
	dDriver.iobase3 = pci_resource_start (pci_device, 3);
	dDriver.iobase4 = pci_resource_start (pci_device, 4);
	dDriver.iobase5 = pci_resource_start (pci_device, 5);
	dDriver.irq = pci_device ->irq;
	request_region (dDriver.iobase0, 64, DEVICE_NAME);
	result = register_chrdev (major, DEVICE_NAME, &advanet_2610_fops);
	if (result < 0)	{
		unregister_chrdev (major, DEVICE_NAME);
		return - EIO;
	}
	if (major == 0)
	  	dDriver.major = result;
	else
		dDriver.major = major;
	init_waitqueue_head (&dDriver.queue);		/* add */
	init_2610_board();

	printk ("iobase0=%x ", dDriver.iobase0);
	printk ("iobase1=%x ", dDriver.iobase1);
	printk ("iobase2=%x ", dDriver.iobase2);
	printk ("iobase3=%x ", dDriver.iobase3);
	printk ("iobase4=%x ", dDriver.iobase4);
	printk ("iobase5=%x ", dDriver.iobase5);
	printk ("irq=%d \n", dDriver.irq);

	printk ("init_module/advanet_2610/end/\n");
	return 0;
}

/* cleanup *******************************************************************/
void cleanup_module (void)
{
	printk ("cleanup_module/advanet_2610/ ");
	release_region (dDriver.iobase0, 64);
	if (kbuf != NULL)	{
		kfree (kbuf);
		kbuf = NULL;
	}
	init_2610_board();
	unregister_chrdev (dDriver.major, DEVICE_NAME);
	printk ("cleanup_module/advanet_2610/end\n");
}
/* end of file ***************************************************************/
